"""Memory."""

from typing import NamedTuple

from psutil import swap_memory, virtual_memory
from psutil._common import sswap

from systembridgeshared.base import Base


class Memory(Base):
    """Memory data."""

    def get_swap(self) -> sswap:
        """Swap memory."""
        return swap_memory()

    def get_virtual(self) -> NamedTuple:
        """Virtual memory."""
        return virtual_memory()



"""Processes."""

from typing import override

from psutil import AccessDenied, NoSuchProcess, process_iter
from systembridgemodels.modules.processes import Process

from systembridgeshared.base import Base


class Processes(Base):
    """Processes data."""

    @override
    async def update_all_data(self) -> list[Process]:
        """Update all data."""
        self._logger.debug("Update all data")

        process_list = list(process_iter())

        # Get names of processes
        items = []
        for process in process_list:
            model = Process(id=process.pid)

            try:
                model.name = process.name()
                model.cpu_usage = process.cpu_percent()
                model.created = process.create_time()
                model.memory_usage = process.memory_percent()
                model.path = process.exe()
                model.status = process.status()
                model.username = process.username()
            except (AccessDenied, NoSuchProcess, OSError):
                pass
            items.append(model)
        # Sort by name
        items = sorted(items, key=lambda item: item.name or "")

        return items



"""System."""

from enum import StrEnum
import getpass
import os
import platform
import re
import socket
import sys
from typing import Any
import uuid

import aiohttp
from packaging.version import parse
from plyer import uniqueid
from psutil import boot_time, users
from psutil._common import suser

from systembridgeshared.base import Base
from systembridgeshared.common import get_user_data_directory

from .._version import __version__


class RunMode(StrEnum):
    """Run Mode."""

    STANDALONE = "standalone"
    PYTHON = "python"


class System(Base):
    """System data."""

    def __init__(self) -> None:
        """Initialise."""
        super().__init__()
        self._mac_address: str = self.get_mac_address()

        # Determine the run mode based on the running executable
        self._run_mode: RunMode = (
            RunMode.PYTHON if "python" in sys.executable.lower() else RunMode.STANDALONE
        )
        self._logger.info("Run mode: %s", self._run_mode)

        # Get the version
        self._version: str | None = None
        if self._run_mode == RunMode.PYTHON:
            self._version = __version__.public()
        if self._run_mode == RunMode.STANDALONE:
            # Read the version file from the package
            with open(
                os.path.join(
                    get_user_data_directory(),
                    "systembridge-version.txt",
                ),
                encoding="utf-8",
            ) as version_file:
                self._version = version_file.read().strip()
        self._logger.info("Version: %s", self._version)

        # Determine the latest version URL based on the run mode
        self._version_latest_url = f"https://github.com/timmo001/{(
            'system-bridge' if self._run_mode == RunMode.STANDALONE else 'system-bridge-backend'
        )}/releases/latest"

        self._version_latest: str | None = None

    def get_active_user_id(self) -> int:
        """Get active user ID."""
        return os.getpid()

    def get_active_user_name(self) -> str | None:
        """Get active user."""
        return getpass.getuser()

    def get_boot_time(self) -> float:
        """Get boot time."""
        return boot_time()

    def get_camera_usage(self) -> list[str]:
        """Return a list of apps that are currently using the webcam."""
        active_apps: list[str] = []
        if sys.platform == "win32":
            # Read from registry for camera usage
            import winreg  # pylint: disable=import-error,import-outside-toplevel

            subkey_path = r"SOFTWARE\Microsoft\Windows\CurrentVersion\CapabilityAccessManager\ConsentStore\webcam"

            def get_subkey_timestamp(subkey) -> int | None:
                """Return the timestamp of the subkey."""
                try:
                    value, _ = winreg.QueryValueEx(subkey, "LastUsedTimeStop")
                    return value
                except OSError:
                    pass
                return None

            try:
                key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, subkey_path)
                # Enumerate over the subkeys of the webcam key
                subkey_count, _, _ = winreg.QueryInfoKey(key)
                # Recursively open each subkey and check the "LastUsedTimeStop" value.
                # A value of 0 means the camera is currently in use.
                for idx in range(subkey_count):
                    subkey_name = winreg.EnumKey(key, idx)
                    subkey_name_full = f"{subkey_path}\\{subkey_name}"
                    subkey = winreg.OpenKey(winreg.HKEY_CURRENT_USER, subkey_name_full)
                    if subkey_name == "NonPackaged":
                        # Enumerate over the subkeys of the "NonPackaged" key
                        subkey_count, _, _ = winreg.QueryInfoKey(subkey)
                        for np_idx in range(subkey_count):
                            subkey_name_np = winreg.EnumKey(subkey, np_idx)
                            subkey_name_full_np = (
                                f"{subkey_path}\\NonPackaged\\{subkey_name_np}"
                            )
                            subkey_np = winreg.OpenKey(
                                winreg.HKEY_CURRENT_USER, subkey_name_full_np
                            )
                            if get_subkey_timestamp(subkey_np) == 0:
                                active_apps.append(subkey_name_np)
                    elif get_subkey_timestamp(subkey) == 0:
                        active_apps.append(subkey_name)
                    winreg.CloseKey(subkey)
                winreg.CloseKey(key)
            except OSError:
                pass
        elif sys.platform in ["darwin", "linux"]:
            # Unknown, please open an issue or PR if you know how to do this
            pass
        return active_apps

    def get_fqdn(self) -> str:
        """Get FQDN."""
        return socket.getfqdn()

    def get_hostname(self) -> str:
        """Get hostname."""
        return socket.gethostname()

    def get_ip_address_4(self) -> str:
        """Get IPv4 address."""
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect(("8.8.8.8", 80))
            return sock.getsockname()[0]
        except OSError:
            return ""

    def get_ip_address_6(self) -> str:
        """Get IPv6 address."""
        try:
            sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
            sock.connect(("2001:4860:4860::8888", 80))
            return sock.getsockname()[0]
        except OSError:
            return ""

    def get_mac_address(self) -> str:
        """Get MAC address."""
        return ":".join(re.findall("..", f"{uuid.getnode():012x}"))

    def get_pending_reboot(self) -> bool:
        """Check if there is a pending reboot."""
        if sys.platform == "win32":
            # Read from registry for pending reboot
            import winreg  # pylint: disable=import-error,import-outside-toplevel

            reg = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
            # Check for "Reboot Required" keys
            try:
                key = winreg.OpenKey(
                    reg,
                    r"SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\RebootRequired",
                )
                winreg.CloseKey(key)
                return True
            except OSError:
                pass
            try:
                key = winreg.OpenKey(
                    reg,
                    r"SOFTWARE\Microsoft\Windows\CurrentVersion\Component Based Servicing\RebootPending",
                )
                winreg.CloseKey(key)
                return True
            except OSError:
                pass
            # Check for recent installation requiring reboot
            try:
                key = winreg.OpenKey(
                    reg,
                    r"SOFTWARE\Microsoft\Updates\UpdateExeVolatile",
                )
                winreg.CloseKey(key)
                return True
            except OSError:
                pass
            # Check for System Center Configuration Manager
            try:
                key = winreg.OpenKey(
                    reg,
                    r"SOFTWARE\Microsoft\SMS\Mobile Client\Reboot Management\RebootData",
                )
                winreg.CloseKey(key)
                return True
            except OSError:
                pass
            # Check for pending file rename operations
            try:
                key = winreg.OpenKey(
                    reg,
                    r"SYSTEM\CurrentControlSet\Control\Session Manager",
                )
                value, _ = winreg.QueryValueEx(key, "PendingFileRenameOperations")
                winreg.CloseKey(key)
                if value:
                    return True
            except OSError:
                pass
        elif sys.platform in ["darwin", "linux"]:
            if os.path.exists("/var/run/reboot-required"):
                return True
            if os.path.exists("/var/run/reboot-required.pkgs"):
                return True
        return False

    def get_platform(self) -> str:
        """Get platform."""
        return platform.system()

    def get_platform_version(self) -> str:
        """Get platform version."""
        return platform.version()

    def get_uptime(self) -> float:
        """Get uptime."""
        return os.times().system

    def get_users(self) -> list[suser]:  # pylint: disable=unsubscriptable-object
        """Get users."""
        return users()

    @property
    def _uuid(self) -> str:
        """Get UUID."""
        # cat /var/lib/dbus/machine-id
        if sys.platform == "linux":
            try:
                with open(
                    "/var/lib/dbus/machine-id",
                    encoding="utf8",
                ) as file:
                    return file.read().strip()
            except FileNotFoundError:
                return self._mac_address

        try:
            return uniqueid.id or self._mac_address
        except Exception:  # pylint: disable=broad-except
            return self._mac_address

    async def _check_rate_limit(self) -> int:
        """Check the GitHub API rate limit."""
        async with aiohttp.ClientSession() as session, session.get(
            "https://api.github.com/rate_limit"
        ) as response:
            if response.status == 200:
                data = await response.json()
                rate_limit = data.get("rate", {})
                return rate_limit.get("remaining", 0)
        return 0

    async def get_version_latest(self) -> Any | None:
        """Get latest version from GitHub."""
        self._logger.info("Get latest version from GitHub")

        # Check if the rate limit allows the request
        rate_limit_remaining = await self._check_rate_limit()
        self._logger.debug("Rate limit: %s", rate_limit_remaining)
        if rate_limit_remaining < 1:
            self._logger.warning("Rate limit exceeded. Skipping request.")
            return self._version_latest

        url = f"https://api.github.com/repos/timmo001/{(
            'system-bridge' if self._run_mode == RunMode.STANDALONE else 'system-bridge-backend'
        )}/releases/latest"
        self._logger.debug("GitHub API URL: %s", url)

        # Use the GitHub API to get the latest release
        async with aiohttp.ClientSession() as session, session.get(url) as response:
            if response.status == 200:
                data = await response.json()
                if data is not None and (tag_name := data.get("tag_name")) is not None:
                    self._version_latest = tag_name.replace("v", "")
                    self._logger.info("Latest version: %s", self._version_latest)

        return self._version_latest

    def get_version_newer_available(self) -> bool | None:
        """Check if newer version is available."""
        if self._version_latest is not None and self._version is not None:
            return parse(self._version_latest) > parse(self._version)
        return None



"""
Provides systembridgedata version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update systembridgedata` to change this file.

from incremental import Version

__version__ = Version("systembridgedata", 5, 0, 0, dev=1)
__all__ = ["__version__"]



"""Disks."""

from psutil import disk_io_counters, disk_partitions, disk_usage
from psutil._common import sdiskio, sdiskpart, sdiskusage

from systembridgeshared.base import Base


class Disks(Base):
    """Disks data."""

    def get_io_counters(self) -> sdiskio | None:
        """Disk IO counters."""
        return disk_io_counters()

    def get_io_counters_per_disk(self) -> dict[str, sdiskio]:
        """Disk IO counters per disk."""
        return disk_io_counters(perdisk=True)

    def get_partitions(self) -> list[sdiskpart]:
        """Disk partitions."""
        return disk_partitions(all=True)

    def get_usage(self, path: str) -> sdiskusage | None:
        """Disk usage."""
        try:
            return disk_usage(path)
        except (FileNotFoundError, PermissionError) as error:
            self._logger.warning(
                "Error getting disk usage for: %s",
                path,
                exc_info=error,
            )
            return None



"""Network."""

from psutil import net_connections, net_if_addrs, net_if_stats, net_io_counters
from psutil._common import sconn, snetio, snicaddr, snicstats

from systembridgeshared.base import Base


class Networks(Base):
    """Networks data."""

    def get_addresses(
        self,
    ) -> dict[str, list[snicaddr]]:
        """Addresses."""
        return net_if_addrs()

    def get_connections(self) -> list[sconn]:
        """Get connections."""
        return net_connections("all")

    def get_io_counters(self) -> snetio:
        """IO Counters."""
        return net_io_counters()

    def get_stats(self) -> dict[str, snicstats]:
        """Stats."""
        return net_if_stats()



"""Media."""

# pylint: disable=import-error
import asyncio
from collections.abc import Awaitable, Callable
import datetime
import platform
from typing import Final

from systembridgemodels.modules.media import Media as MediaInfo
from winsdk.windows.foundation import EventRegistrationToken
import winsdk.windows.media.control as wmc

from systembridgeshared.base import Base

IDLE_UPDATE_INTERVAL: Final[int] = 20
PLAYING_UPDATE_INTERVAL: Final[int] = 5


class Media(Base):
    """Media data."""

    def __init__(
        self,
        changed_callback: Callable[[str, MediaInfo], Awaitable[None]],
        update_media_info_interval: Callable[[int], None],
    ) -> None:
        """Initialise."""
        super().__init__()
        self._changed_callback = changed_callback

        self.sessions: None | (
            wmc.GlobalSystemMediaTransportControlsSessionManager
        ) = None
        self.current_session: None | (
            wmc.GlobalSystemMediaTransportControlsSession
        ) = None
        self.current_session_changed_handler_token: None | (
            EventRegistrationToken
        ) = None
        self.properties_changed_handler_token: EventRegistrationToken | None = None
        self.playback_info_changed_handler_token: None | (EventRegistrationToken) = None

        self.update_media_info_interval = update_media_info_interval

    def _current_session_changed_handler(
        self,
        _sender,
        _result,
    ) -> None:
        """Session changed handler."""
        self._logger.info("Session changed")
        if self._changed_callback is not None:
            asyncio.run(self.update_media_info())

    def _properties_changed_handler(
        self,
        _sender,
        _result,
    ) -> None:
        """Properties changed handler."""
        self._logger.info("Media properties changed")
        if self._changed_callback is not None:
            asyncio.run(self.update_media_info())

    def _playback_info_changed_handler(
        self,
        _sender,
        _result,
    ) -> None:
        """Playback info changed handler."""
        self._logger.info("Media properties changed")
        if self._changed_callback is not None:
            asyncio.run(self.update_media_info())

    async def _update_data(
        self,
        media_info: MediaInfo | None = None,
    ) -> None:
        """Update data."""
        if media_info is None:
            media_info = MediaInfo(updated_at=datetime.datetime.now().timestamp())

        self._logger.info("Updating media data")
        await self._changed_callback("media", media_info)

    async def update_media_info(self) -> None:
        """Update media info from the current session."""
        if platform.system() != "Windows":
            return None

        if (
            self.sessions is not None
            and self.current_session_changed_handler_token is not None
        ):
            self.sessions.remove_current_session_changed(
                self.current_session_changed_handler_token
            )

        self.sessions = (
            await wmc.GlobalSystemMediaTransportControlsSessionManager.request_async()
        )
        self.current_session_changed_handler_token = (
            self.sessions.add_current_session_changed(
                self._current_session_changed_handler
            )
        )

        if self.current_session is not None:
            if self.properties_changed_handler_token is not None:
                self.current_session.remove_media_properties_changed(
                    self.properties_changed_handler_token
                )
            if self.playback_info_changed_handler_token is not None:
                self.current_session.remove_playback_info_changed(
                    self.playback_info_changed_handler_token
                )

        self.current_session = self.sessions.get_current_session()
        if self.current_session:
            self.properties_changed_handler_token = (
                self.current_session.add_media_properties_changed(
                    self._properties_changed_handler
                )
            )
            self.playback_info_changed_handler_token = (
                self.current_session.add_playback_info_changed(
                    self._playback_info_changed_handler
                )
            )
            media_info = MediaInfo()
            if info := self.current_session.get_playback_info():
                media_info.status = info.playback_status.name
                media_info.playback_rate = info.playback_rate
                media_info.shuffle = info.is_shuffle_active
                if info.auto_repeat_mode:
                    media_info.repeat = info.auto_repeat_mode.name
                if info.playback_type:
                    media_info.type = info.playback_type.name
                if info.controls:
                    media_info.is_fast_forward_enabled = (
                        info.controls.is_fast_forward_enabled
                    )
                    media_info.is_next_enabled = info.controls.is_next_enabled
                    media_info.is_pause_enabled = info.controls.is_pause_enabled
                    media_info.is_play_enabled = info.controls.is_play_enabled
                    media_info.is_previous_enabled = info.controls.is_previous_enabled
                    media_info.is_rewind_enabled = info.controls.is_rewind_enabled
                    media_info.is_stop_enabled = info.controls.is_stop_enabled

            if timeline := self.current_session.get_timeline_properties():
                media_info.duration = timeline.end_time.total_seconds()
                media_info.position = timeline.position.total_seconds()

            if (
                properties
                := await self.current_session.try_get_media_properties_async()
            ):
                media_info.title = properties.title
                media_info.subtitle = properties.subtitle
                media_info.artist = properties.artist
                media_info.album_artist = properties.album_artist
                media_info.album_title = properties.album_title
                media_info.track_number = properties.track_number

            media_info.updated_at = datetime.datetime.now().timestamp()

            await self._update_data(media_info)

            if media_info.status == "PLAYING":
                self.update_media_info_interval(PLAYING_UPDATE_INTERVAL)
            else:
                self.update_media_info_interval(IDLE_UPDATE_INTERVAL)
        else:
            await self._update_data(
                MediaInfo(updated_at=datetime.datetime.now().timestamp())
            )



"""CPU."""

from psutil import (
    cpu_count,
    cpu_freq,
    cpu_percent,
    cpu_stats,
    cpu_times,
    cpu_times_percent,
    getloadavg,
)
from psutil._common import pcputimes, scpufreq, scpustats, shwtemp
from systembridgemodels.modules.sensors import Sensors

from systembridgeshared.base import Base


class CPU(Base):
    """CPU data."""

    def __init__(self) -> None:
        """Initialise."""
        super().__init__()

        self._count: int = cpu_count()

        self.sensors: Sensors | None = None

    def get_frequency(self) -> scpufreq:
        """CPU frequency."""
        return cpu_freq()

    def get_frequency_per_cpu(
        self,
    ) -> list[scpufreq]:
        """CPU frequency per CPU."""
        return cpu_freq(percpu=True)  # type: ignore

    def get_load_average(self) -> float:
        """Get load average."""
        avg_tuple = getloadavg()
        return sum([avg_tuple[0], avg_tuple[1], avg_tuple[2]]) / 3

    def get_power_package(self) -> float | None:
        """CPU package power."""
        if (
            self.sensors is None
            or self.sensors.windows_sensors is None
            or self.sensors.windows_sensors.hardware is None
        ):
            return None
        for hardware in self.sensors.windows_sensors.hardware:
            # Find type "CPU"
            if "CPU" not in hardware.type.upper():
                continue
            for sensor in hardware.sensors:
                # Find type "POWER" and name "PACKAGE"
                if (
                    "POWER" in sensor.type.upper()
                    and "PACKAGE" in sensor.name.upper()
                    and sensor.value is not None
                ):
                    self._logger.debug(
                        "Found CPU package power: %s = %s", sensor.name, sensor.value
                    )
                    return (
                        float(sensor.value)
                        if isinstance(sensor.value, (int, float))
                        else None
                    )
        return None

    def get_power_per_cpu(self) -> list[float] | None:
        """CPU package power."""
        powers: list[float] = [-1] * self._count
        if (
            self.sensors is None
            or self.sensors.windows_sensors is None
            or self.sensors.windows_sensors.hardware is None
        ):
            return None
        for hardware in self.sensors.windows_sensors.hardware:
            # Find type "CPU"
            if "CPU" not in hardware.type.upper():
                continue
            for sensor in hardware.sensors:
                # Find type "POWER" and name "CORE"
                if (
                    "POWER" in sensor.type.upper()
                    and "CORE" in sensor.name.upper()
                    and sensor.value is not None
                ):
                    self._logger.debug(
                        "Found CPU core power: %s (%s) = %s",
                        sensor.name,
                        sensor.id,
                        sensor.value,
                    )
                    for sensor in hardware.sensors:
                        # Find type "POWER" and name "PACKAGE"
                        if (
                            "POWER" in sensor.type.upper()
                            and "PACKAGE" not in sensor.name.upper()
                            and sensor.value is not None
                        ):
                            self._logger.debug(
                                "Found CPU package power: %s (%s) = %s",
                                sensor.name,
                                sensor.id,
                                sensor.value,
                            )
                            index = int(sensor.id.split("/")[-1])
                            powers[index] = float(sensor.value)

        return powers

    def get_stats(self) -> scpustats:
        """CPU stats."""
        return cpu_stats()

    def get_temperature(self) -> float | None:
        """CPU temperature."""
        if self.sensors is not None:
            if self.sensors.temperatures is not None:
                temperatures: dict[str, list[shwtemp]] = self.sensors.temperatures
                if "k10temp" in temperatures:
                    for sensor in self.sensors.temperatures["k10temp"]:
                        self._logger.debug("k10temp: %s", sensor)
                        if "Tdie" in sensor or "Tctl" in sensor or "Tccd1" in sensor:
                            self._logger.debug(
                                "Found CPU temperature (k10temp): %s",
                                sensor,
                            )
                            return sensor.current
                if "coretemp" in self.sensors.temperatures:
                    for sensor in self.sensors.temperatures["coretemp"]:
                        self._logger.debug("coretemp: %s", sensor)
                        if (
                            "Package id 0" in sensor
                            or "Physical id 0" in sensor
                            or "Core 0" in sensor
                        ):
                            self._logger.debug(
                                "Found CPU temperature (coretemp): %s",
                                sensor,
                            )
                            return sensor.current
                if "atk0110" in self.sensors.temperatures:
                    for sensor in self.sensors.temperatures["atk0110"]:
                        self._logger.debug("atk0110: %s", sensor)
                        if "CPU" in sensor:
                            self._logger.debug(
                                "Found CPU temperature (atk0110): %s",
                                sensor,
                            )
                            return sensor.current
                # Use the first temperature sensor key
                for key in temperatures:
                    for sensor in temperatures[key]:
                        self._logger.warning(
                            "Unknown sensor used (may not be correct): %s", sensor
                        )
                        return sensor.current
            if (
                self.sensors.windows_sensors is not None
                and self.sensors.windows_sensors.hardware is not None
            ):
                for hardware in self.sensors.windows_sensors.hardware:
                    # Find type "CPU"
                    if "CPU" not in hardware.type.upper():
                        continue
                    for sensor in hardware.sensors:
                        name = sensor.name.upper()
                        # Find type "TEMPERATURE" and name "PACKAGE" or "AVERAGE"
                        if (
                            "TEMPERATURE" in sensor.type.upper()
                            and ("PACKAGE" in name or "AVERAGE" in name)
                            and sensor.value is not None
                        ):
                            self._logger.debug(
                                "Found CPU temperature: %s = %s",
                                sensor.name,
                                sensor.value,
                            )
                            return (
                                float(sensor.value)
                                if isinstance(sensor.value, (int, float, str))
                                else None
                            )
        return None

    def get_times(self) -> pcputimes:
        """CPU times."""
        return cpu_times(percpu=False)

    def get_times_percent(self) -> pcputimes:
        """CPU times percent."""
        return cpu_times_percent(interval=1, percpu=False)

    def get_times_per_cpu(
        self,
    ) -> list[pcputimes]:
        """CPU times per CPU."""
        return cpu_times(percpu=True)

    def get_times_per_cpu_percent(
        self,
    ) -> list[pcputimes]:
        """CPU times per CPU percent."""
        return cpu_times_percent(interval=1, percpu=True)

    def get_usage(self) -> float:
        """CPU usage."""
        return cpu_percent(interval=1, percpu=False)

    def get_usage_per_cpu(
        self,
    ) -> list[float]:
        """CPU usage per CPU."""
        return cpu_percent(interval=1, percpu=True)  # type: ignore

    def get_voltages(self) -> tuple[float | None, list[float]]:
        """CPU voltage."""
        voltage: float | None = None
        voltages: list[float] = [-1] * self._count
        voltage_sensors = []
        if (
            self.sensors is None
            or self.sensors.windows_sensors is None
            or self.sensors.windows_sensors.hardware is None
        ):
            return (voltage, voltages)
        for hardware in self.sensors.windows_sensors.hardware:
            # Find type "CPU"
            if "CPU" not in hardware.type.upper():
                continue
            for sensor in hardware.sensors:
                # Find type "VOLTAGE"
                if "VOLTAGE" in sensor.type.upper() and sensor.value is not None:
                    self._logger.debug(
                        "Found CPU voltage: %s (%s) = %s",
                        sensor.name,
                        sensor.id,
                        sensor.value,
                    )
                    voltage_sensors.append(sensor)

        # Handle voltages
        if voltage_sensors is not None and len(voltage_sensors) > 0:
            for sensor in voltage_sensors:
                if sensor.value is not None:
                    self._logger.debug(
                        "CPU voltage: %s (%s) = %s",
                        sensor.name,
                        sensor.type,
                        sensor.value,
                    )
                    # "/amdcpu/0/voltage/16" -> 16
                    # Get the last part of the id
                    index = int(sensor.id.split("/")[-1])
                    if 0 <= index < self._count:
                        voltages[index] = float(sensor.value)
            voltage_sum = 0
            for voltage in voltages:
                if voltage is not None:
                    voltage_sum += voltage
            if voltage_sum > 0:
                voltage = voltage_sum / self._count
            else:
                # If we can't get the average, just use the first value
                voltage = voltage_sensors[0].value

        return (voltage, voltages)
